import math
from typing import Tuple
import glglue.gltf


def translation(t) -> str:
    return f'position [{t[0]:.2f}, {t[1]:.2f}, {t[2]:.2f}]'


def inverseMatrix(m) -> str:
    return f'[{m[12]:.2f}, {m[13]:.2f}, {m[14]:.2f}, {m[15]:.2f}]'


def get_world_position(data: glglue.gltf.GltfData, node: glglue.gltf.GltfNode) -> Tuple[float, float, float]:
    # TODO:
    return 0, 0, 0


def info(data: glglue.gltf.GltfData, skin_index: int):
    skin = data.gltf['skins'][skin_index]
    joints = skin['joints']

    matrices_index = skin['inverseBindMatrices']
    matrices = data.buffer_reader.read_accessor(matrices_index)
    if matrices.get_count() != len(joints):
        raise Exception()

    text = ''
    for i, joint in enumerate(joints):
        # node = gltf['nodes'][joint]
        node = data.nodes[joint]
        matrix = matrices.get_item(i)

        validation = ''
        EPS = 1e-5
        world_position = get_world_position(data, node)
        if math.fabs(world_position[0] + matrix[12]) > EPS:
            validation += 'X'
        if math.fabs(world_position[1] + matrix[13]) > EPS:
            validation += 'Y'
        if math.fabs(world_position[2] + matrix[14]) > EPS:
            validation += 'Z'

        text += f'{joint}: {translation(world_position)}, {inverseMatrix(matrix)}{validation}\n'
    return text



from PySide6.QtCore import Qt, QAbstractItemModel, QModelIndex
from typing import Optional


class Item:
    def __init__(self, value=None):
        self.name = ''
        self.value = value
        self.parentItem: Optional['Item'] = None
        self.children = []

    def columnCount(self) -> int:
        return 2

    def add_child(self, child):
        self.children.append(child)
        child.parentItem = self

    def childCount(self) -> int:
        return len(self.children)

    def child(self, row):
        return self.children[row]

    def data(self, column):
        if column == 0:
            return self.name
        elif column == 1:
            return self.value
        else:
            raise Exception()

    def parent(self):
        return self.parentItem

    def row(self):
        if isinstance(self.parentItem, Item):
            return self.parentItem.children.index(self)
        return 0

    def get_ancestors(self):
        yield self
        if self.parentItem:
            for x in self.parentItem.get_ancestors():
                yield x

    def json_path(self) -> str:
        path = [f'/{x.name}' for x in self.get_ancestors()][:-1]
        return ''.join(reversed(path))


def build(value):
    if isinstance(value, dict):
        node = Item()
        for k, v in value.items():
            child = build(v)
            node.add_child(child)
            child.name = k
        return node

    elif isinstance(value, list):
        node = Item()
        for i, v in enumerate(value):
            child = build(v)
            node.add_child(child)
            child.name = f'{i}'
        return node

    else:
        node = Item(value)
        return node


class TreeModel(QAbstractItemModel):
    def __init__(self, root: dict, parent=None):
        super(TreeModel, self).__init__(parent)
        self.rootItem = build(root)
        # self.layoutChanged.emit()

    def columnCount(self, parent):
        if parent.isValid():
            return parent.internalPointer().columnCount()
        else:
            return self.rootItem.columnCount()

    def data(self, index, role):

        if not index.isValid():
            return None
        if role != Qt.DisplayRole:
            return None
        item = index.internalPointer()
        return item.data(index.column())

    def flags(self, index):

        if not index.isValid():
            return Qt.NoItemFlags
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def headerData(self, section, orientation, role):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            # return self.rootItem.data(section)
            if section == 0:
                return 'property'
            elif section == 1:
                return 'value'
            else:
                raise Exception()

        return None

    def index(self, row, column, parent):

        if not parent.isValid():
            parentItem = self.rootItem
        else:
            parentItem = parent.internalPointer()
        childItem = parentItem.child(row)
        if childItem:
            return self.createIndex(row, column, childItem)
        else:
            return QModelIndex()

    def parent(self, index: QModelIndex) -> QModelIndex:

        if not index.isValid():
            return QModelIndex()
        childItem = index.internalPointer()
        parentItem = childItem.parent()
        if parentItem == self.rootItem:
            return QModelIndex()
        return self.createIndex(parentItem.row(), 0, parentItem)

    def rowCount(self, parent):

        if parent.column() > 0:
            return 0
        if not parent.isValid():
            parentItem = self.rootItem
        else:
            parentItem = parent.internalPointer()
        return parentItem.childCount()



from . import json_tree
from PySide6 import QtWidgets, QtCore, QtGui
from typing import Optional, Union
import re
import pathlib
from logging import getLogger
logger = getLogger(__name__)


class Window(QtWidgets.QMainWindow):
    '''
    +----+----+----+
    |json|Open|Prop|
    |tree|GL  |    |
    +----+----+----+
    '''

    def __init__(self, parent=None):
        # import glglue.PySide6gl
        super().__init__(parent)

        self.resize(1280, 1024)
        self.create_menu()

        # setup opengl widget
        import glglue.gl3.samplecontroller
        self.controller = glglue.gl3.samplecontroller.SampleController()
        import glglue.pyside6
        import glglue.utils
        self.glwidget = glglue.pyside6.Widget(
            self, self.controller, glglue.utils.get_desktop_scaling_factor())
        self.setCentralWidget(self.glwidget)

        # left json tree
        self.dock_left = QtWidgets.QDockWidget("json", self)
        self.addDockWidget(QtGui.Qt.LeftDockWidgetArea, self.dock_left)
        self.json_tree = QtWidgets.QTreeView(self.dock_left)
        self.dock_left.setWidget(self.json_tree)

        # right selected panel
        self.dock_right = QtWidgets.QDockWidget("active", self)
        self.addDockWidget(QtGui.Qt.RightDockWidgetArea, self.dock_right)
        self.text = QtWidgets.QTextEdit(self)
        self.dock_right.setWidget(self.text)

        # logger
        import logging
        self.logger = glglue.pyside6.QPlainTextEditLogger(self)
        logging.getLogger('').addHandler(self.logger.log_handler)
        self.dock_bottom = QtWidgets.QDockWidget("logger", self)
        self.addDockWidget(QtGui.Qt.BottomDockWidgetArea, self.dock_bottom)
        self.dock_bottom.setWidget(self.logger)

    def create_menu(self):
        mainMenu = self.menuBar()
        fileMenu = mainMenu.addMenu("File")
        # viewMenu = mainMenu.addMenu("View")
        # editMenu = mainMenu.addMenu("Edit")
        # searchMenu = mainMenu.addMenu("Font")
        # helpMenu = mainMenu.addMenu("Help")

        openAction = QtGui.QAction(QtGui.QIcon('open.png'), "Open", self)
        openAction.setShortcut("Ctrl+O")
        openAction.triggered.connect(self.open_dialog)  # type: ignore
        fileMenu.addAction(openAction)

        # saveAction = QAction(QIcon('save.png'), "Save", self)
        # saveAction.setShortcut("Ctrl+S")
        # fileMenu.addAction(saveAction)

        exitAction = QtGui.QAction(QtGui.QIcon('exit.png'), "Exit", self)
        exitAction.setShortcut("Ctrl+X")
        exitAction.triggered.connect(self.exit_app)  # type: ignore
        fileMenu.addAction(exitAction)

    def exit_app(self):
        self.close()

    def open_dialog(self):
        dlg = QtWidgets.QFileDialog()
        dlg.setFileMode(QtWidgets.QFileDialog.AnyFile)
        dlg.setFilter(QtCore.QDir.Files)
        dlg.setNameFilters(['*.gltf;*.glb;*.vrm;*.vci', '*'])
        if dlg.exec_():
            filenames = dlg.selectedFiles()
            if filenames and len(filenames) > 0:
                self.open(pathlib.Path(filenames[0]))

    def open(self, file: pathlib.Path):
        logger.info(f'load: {file.name}')
        self.setWindowTitle(str(file.name))
        import gltfio
        self.gltf = gltfio.parse_path(file)

        # opengl
        import glglue.gltf_loader
        loader = glglue.gltf_loader.GltfLoader(self.gltf)
        scene = loader.load()
        self.controller.drawables = [scene]
        self.glwidget.repaint()

        # json
        self.open_json(self.gltf.gltf, file, self.gltf.bin)

    def open_json(self, gltf_json: dict, path: pathlib.Path, bin: Optional[bytes]):
        self.gltf_json = gltf_json
        self.json_model = json_tree.TreeModel(gltf_json)
        self.json_tree.setModel(self.json_model)
        self.json_tree.selectionModel().selectionChanged.connect(  # type: ignore
            self.on_selected)
        self.bin = bin

    def on_selected(self, selected, deselected):
        selected = selected.indexes()
        if not selected:
            return
        item = selected[0].internalPointer()
        if not isinstance(item, json_tree.Item):
            return

        json_path = item.json_path()
        m = re.match(r'^/skins/(\d+)$', json_path)
        if self.gltf and m:
            groups = m.groups()
            skin_index = int(groups[0])
            from . import skin_debug
            self.text.setText(
                skin_debug.info(self.gltf, skin_index))
        else:
            self.text.setText(json_path)


def run():
    import sys
    from logging import basicConfig, DEBUG
    basicConfig(format='%(levelname)s:%(name)s:%(message)s', level=DEBUG)
    app = QtWidgets.QApplication(sys.argv)
    window = Window()
    if len(sys.argv) > 1:
        window.open(pathlib.Path(sys.argv[1]))
    window.show()
    sys.exit(app.exec_())
